unit Keyboard;

interface

type

  keyblock = record
    L: integer;		{ left }
    T: integer;		{ top }
    W: integer;		{ width of the key }
    H: integer;		{ height of the key }
    SX: integer;	{ horizontal spacing }
    SY: integer;	{ vertical spacing }
    col: integer;	{ number of columns }
    cnt: integer;	{ number of keys in a block }
    OX: integer;	{ left in the Keys.bmp or 512 if absent }
    OY: integer;	{ top in the Keys.bmp or 512 if absent }
  end;


const

  KEYPADS = 12;		{index of the last item in the 'keypad' array}
  LASTKEYCODE = 99;

  keypad: array[0..KEYPADS] of keyblock = (
{ power switch, code:1 }
    (	L:34;	T:350;	W:46;	H:16;	SX:46;	SY:35;	col:1;	cnt:1;	OX:128;	OY:17	),
{ cover lock, code:2 }
    (	L:465;	T:1;	W:113;	H:47;	SX:113;	SY:47;	col:1;	cnt:1;	OX:0;	OY:38	),
{ New All and Reset keys, code: 3..4 }
    (	L:330;	T:365;	W:5;	H:6;	SX:40;	SY:35;	col:2;	cnt:2;	OX:220;	OY:17	),
{ first two rows of small keys: F to Contrast and Q to P, code: 5..24 }
    (	L:25;	T:394;	W:27;	H:17;	SX:37;	SY:35;	col:10;	cnt:20;	OX:129;	OY:0	),
{ third row of small keys: A to L, code: 25..33 }
    (	L:36;	T:464;	W:27;	H:17;	SX:37;	SY:35;	col:9;	cnt:9;	OX:129;	OY:0	),
{ fourth row of small keys: Caps to Eng, code: 34..42 }
    (	L:25;	T:499;	W:27;	H:17;	SX:37;	SY:35;	col:9;	cnt:9;	OX:129;	OY:0	),
{ SHIFT keys, code: 43..44 }
    (	L:25;	T:534;	W:45;	H:17;	SX:315;	SY:35;	col:2;	cnt:2;	OX:111;	OY:0	),
{ SPACE bar, code: 45 }
    (	L:127;	T:534;	W:156;	H:17;	SX:156;	SY:35;	col:1;	cnt:1;	OX:0;	OY:0	),
{ five rows of large keys, code: 46..68 }
    (	L:403;	T:366;	W:27;	H:21;	SX:37;	SY:41;	col:5;	cnt:23;	OX:37;	OY:17	),
{ EXE key, code: 69 }
    (	L:514;	T:530;	W:64;	H:21;	SX:64;	SY:41;	col:1;	cnt:1;	OX:0;	OY:17	),
{ row of 13 membrane keys below the LCD, code: 70..82 }
    (	L:62;	T:239;	W:36;	H:30;	SX:38;	SY:30;	col:13;	cnt:13;	OX:512;	OY:512	),
{ touch screen pads, code: 83..98 }
    (	L:74;	T:106;	W:96;	H:16;	SX:96;	SY:16;	col:4;	cnt:16;	OX:512;	OY:512	),
{ KANA key, code 99 }
    (	L:358;	T:499;	W:27;	H:17;	SX:37;	SY:35;	col:1;	cnt:1;	OX:129;	OY:0	)
  );


var
  KeyCode1: integer;		{ from the mouse }
  KeyCode2: integer;		{ from the keyboard }


  function ReadKy (Ko: byte) : word;


implementation

const

{ tables converting KeyCode1 and KeyCode2 to the KY state for given KO }

  KeyTab: array[0..15, 0..LASTKEYCODE] of word = (

( { KO code 0, none selected }
{ no key pressed }
  $0000,
{ power switch, cover lock, New All and Reset keys, code: 1..4 }
  $0000, $0000, $0000, $0000,
{ first row of small keys: F to Contrast, code: 5..14 }
  $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000,
{ second row of small keys: Q to P, code: 15..24 }
  $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000,
{ third row of small keys: A to L, code: 25..33 }
  $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000,
{ fourth row of small keys: Caps to Eng, code: 34..42 }
  $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000,
{ two SHIFT keys and the SPACE bar, code: 43..45 }
  $0000, $0000, $0000,
{ five rows of large keys, code: 46..68 }
  $0000, $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000, $0000,
  $0000, $0000, $0000,
{ EXE key, code: 69 }
  $0000,
{ row of 13 membrane keys below the LCD, code: 70..82 }
  $0000, $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000, $0000,
  $0000, $0000, $0000,
{ touch screen pads, code: 83..98 }
  $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000,
{ KANA key, code 99 }
  $0000
),

( { KO1 }
{ no key pressed }
  $0000,
{ power switch, cover lock, New All and Reset keys, code: 1..4 }
  $0000, $0000, $0000, $0000,
{ first row of small keys: F to Contrast, code: 5..14 }
  $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000,
{ second row of small keys: Q to P, code: 15..24 }
  $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000,
{ third row of small keys: A to L, code: 25..33 }
  $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000,
{ fourth row of small keys: Caps to Eng, code: 34..42 }
  $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000,
{ two SHIFT keys and the SPACE bar, code: 43..45 }
  $0000, $0000, $0000,
{ five rows of large keys, code: 46..68 }
  $0080, $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000, $0000,
  $0000, $0000, $0000,
{ EXE key, code: 69 }
  $0000,
{ row of 13 membrane keys below the LCD, code: 70..82 }
  $0000, $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000, $0000,
  $0000, $0000, $0000,
{ touch screen pads, code: 83..98 }
  $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000,
{ KANA key, code 99 }
  $0000
),

( { KO2 }
{ no key pressed }
  $0000,
{ power switch, cover lock, New All and Reset keys, code: 1..4 }
  $0000, $0000, $0000, $0000,
{ first row of small keys: F to Contrast, code: 5..14 }
  $0000, $0020, $0010, $0008, $0004, $0002, $0001, $0080, $0000, $0000,
{ second row of small keys: Q to P, code: 15..24 }
  $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000,
{ third row of small keys: A to L, code: 25..33 }
  $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000,
{ fourth row of small keys: Caps to Eng, code: 34..42 }
  $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000,
{ two SHIFT keys and the SPACE bar, code: 43..45 }
  $0000, $0000, $0000,
{ five rows of large keys, code: 46..68 }
  $0000, $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000, $0000,
  $0000, $0000, $0000,
{ EXE key, code: 69 }
  $0000,
{ row of 13 membrane keys below the LCD, code: 70..82 }
  $0000, $0000, $0000, $0000, $2000,
  $4000, $0000, $0000, $0000, $0000,
  $0000, $0000, $8000,
{ touch screen pads, code: 83..98 }
  $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000,
{ KANA key, code 99 }
  $0000
),

( { KO3 }
{ no key pressed }
  $0000,
{ power switch, cover lock, New All and Reset keys, code: 1..4 }
  $0000, $0000, $0000, $0000,
{ first row of small keys: F to Contrast, code: 5..14 }
  $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000,
{ second row of small keys: Q to P, code: 15..24 }
  $0020, $0010, $0008, $0004, $0002, $0001, $0080, $0000, $0000, $0000,
{ third row of small keys: A to L, code: 25..33 }
  $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000,
{ fourth row of small keys: Caps to Eng, code: 34..42 }
  $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000,
{ two SHIFT keys and the SPACE bar, code: 43..45 }
  $0000, $0000, $0000,
{ five rows of large keys, code: 46..68 }
  $0000, $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000, $0000,
  $0000, $0000, $0000,
{ EXE key, code: 69 }
  $0000,
{ row of 13 membrane keys below the LCD, code: 70..82 }
  $0000, $0000, $0000, $2000, $0000,
  $0000, $4000, $0000, $0000, $8000,
  $0000, $0000, $0000,
{ touch screen pads, code: 83..98 }
  $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000,
{ KANA key, code 99 }
  $0000
),

( { KO4 }
{ no key pressed }
  $0000,
{ power switch, cover lock, New All and Reset keys, code: 1..4 }
  $0000, $0000, $0000, $0000,
{ first row of small keys: F to Contrast, code: 5..14 }
  $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000,
{ second row of small keys: Q to P, code: 15..24 }
  $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000,
{ third row of small keys: A to L, code: 25..33 }
  $0010, $0008, $0004, $0002, $0001, $0080, $0000, $0000, $0000,
{ fourth row of small keys: Caps to Eng, code: 34..42 }
  $0020, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000,
{ two SHIFT keys and the SPACE bar, code: 43..45 }
  $0000, $0000, $0000,
{ five rows of large keys, code: 46..68 }
  $0000, $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000, $0000,
  $0000, $0000, $0000,
{ EXE key, code: 69 }
  $0000,
{ row of 13 membrane keys below the LCD, code: 70..82 }
  $0000, $0000, $2000, $0000, $0000,
  $0000, $0000, $4000, $0000, $0000,
  $0000, $8000, $0000,
{ touch screen pads, code: 83..98 }
  $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000,
{ KANA key, code 99 }
  $0000
),

( { KO5 }
{ no key pressed }
  $0000,
{ power switch, cover lock, New All and Reset keys, code: 1..4 }
  $0000, $0000, $0000, $0000,
{ first row of small keys: F to Contrast, code: 5..14 }
  $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000,
{ second row of small keys: Q to P, code: 15..24 }
  $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000,
{ third row of small keys: A to L, code: 25..33 }
  $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000,
{ fourth row of small keys: Caps to Eng, code: 34..42 }
  $0000, $0020, $0010, $0008, $0004, $0002, $0001, $0080, $0000,
{ two SHIFT keys and the SPACE bar, code: 43..45 }
  $0000, $0000, $0000,
{ five rows of large keys, code: 46..68 }
  $0000, $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000, $0000,
  $0000, $0000, $0000,
{ EXE key, code: 69 }
  $0000,
{ row of 13 membrane keys below the LCD, code: 70..82 }
  $0000, $2000, $0000, $0000, $0000,
  $0000, $0000, $0000, $4000, $0000,
  $8000, $0000, $0000,
{ touch screen pads, code: 83..98 }
  $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000,
{ KANA key, code 99 }
  $0000
),

( { KO6 }
{ no key pressed }
  $0000,
{ power switch, cover lock, New All and Reset keys, code: 1..4 }
  $0000, $0000, $0004, $0000,
{ first row of small keys: F to Contrast, code: 5..14 }
  $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0080,
{ second row of small keys: Q to P, code: 15..24 }
  $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000,
{ third row of small keys: A to L, code: 25..33 }
  $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000,
{ fourth row of small keys: Caps to Eng, code: 34..42 }
  $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000,
{ two SHIFT keys and the SPACE bar, code: 43..45 }
  $0000, $0000, $0000,
{ five rows of large keys, code: 46..68 }
  $0000, $0001, $0002, $0008, $0010,
  $0000, $0000, $0000, $0000, $0020,
  $0000, $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000, $0000,
  $0000, $0000, $0000,
{ EXE key, code: 69 }
  $0000,
{ row of 13 membrane keys below the LCD, code: 70..82 }
  $2000, $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000, $0000,
  $0000, $0000, $0000,
{ touch screen pads, code: 83..98 }
  $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000,
{ KANA key, code 99 }
  $0000
),

( { KO7 }
{ no key pressed }
  $0000,
{ power switch, cover lock, New All and Reset keys, code: 1..4 }
  $0000, $0000, $0000, $0000,
{ first row of small keys: F to Contrast, code: 5..14 }
  $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0080, $0000,
{ second row of small keys: Q to P, code: 15..24 }
  $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000,
{ third row of small keys: A to L, code: 25..33 }
  $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000,
{ fourth row of small keys: Caps to Eng, code: 34..42 }
  $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0001,
{ two SHIFT keys and the SPACE bar, code: 43..45 }
  $0000, $0000, $0000,
{ five rows of large keys, code: 46..68 }
  $0000, $0000, $0000, $0000, $0000,
  $0002, $0004, $0008, $0010, $0000,
  $0000, $0000, $0000, $0000, $0020,
  $0000, $0000, $0000, $0000, $0000,
  $0000, $0000, $0000,
{ EXE key, code: 69 }
  $0000,
{ row of 13 membrane keys below the LCD, code: 70..82 }
  $0000, $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000, $0000,
  $0000, $0000, $0000,
{ touch screen pads, code: 83..98 }
  $1000, $0000, $0000, $0000,
  $2000, $0000, $0000, $0000,
  $4000, $0000, $0000, $0000,
  $8000, $0000, $0000, $0000,
{ KANA key, code 99 }
  $0000
),

( { KO8 }
{ no key pressed }
  $0000,
{ power switch, cover lock, New All and Reset keys, code: 1..4 }
  $0000, $0000, $0000, $0000,
{ first row of small keys: F to Contrast, code: 5..14 }
  $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000,
{ second row of small keys: Q to P, code: 15..24 }
  $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0080, $0001, $0002,
{ third row of small keys: A to L, code: 25..33 }
  $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000,
{ fourth row of small keys: Caps to Eng, code: 34..42 }
  $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000,
{ two SHIFT keys and the SPACE bar, code: 43..45 }
  $0000, $0000, $0000,
{ five rows of large keys, code: 46..68 }
  $0000, $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000, $0000,
  $0004, $0008, $0010, $0020, $0000,
  $0000, $0000, $0000, $0000, $0000,
  $0000, $0000, $0000,
{ EXE key, code: 69 }
  $0000,
{ row of 13 membrane keys below the LCD, code: 70..82 }
  $0000, $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000, $0000,
  $0000, $0000, $0000,
{ touch screen pads, code: 83..98 }
  $0000, $1000, $0000, $0000,
  $0000, $2000, $0000, $0000,
  $0000, $4000, $0000, $0000,
  $0000, $8000, $0000, $0000,
{ KANA key, code 99 }
  $0000
),

( { KO9 }
{ no key pressed }
  $0000,
{ power switch, cover lock, New All and Reset keys, code: 1..4 }
  $0000, $0000, $0000, $0000,
{ first row of small keys: F to Contrast, code: 5..14 }
  $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000,
{ second row of small keys: Q to P, code: 15..24 }
  $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000,
{ third row of small keys: A to L, code: 25..33 }
  $0000, $0000, $0000, $0000, $0000, $0000, $0080, $0001, $0002,
{ fourth row of small keys: Caps to Eng, code: 34..42 }
  $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000,
{ two SHIFT keys and the SPACE bar, code: 43..45 }
  $0000, $0000, $0000,
{ five rows of large keys, code: 46..68 }
  $0000, $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000, $0000,
  $0004, $0008, $0010, $0020, $0000,
  $0000, $0000, $0000,
{ EXE key, code: 69 }
  $0000,
{ row of 13 membrane keys below the LCD, code: 70..82 }
  $0000, $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000, $0000,
  $0000, $0000, $0000,
{ touch screen pads, code: 83..98 }
  $0000, $0000, $1000, $0000,
  $0000, $0000, $2000, $0000,
  $0000, $0000, $4000, $0000,
  $0000, $0000, $8000, $0000,
{ KANA key, code 99 }
  $0000
),

( { KO10 }
{ no key pressed }
  $0000,
{ power switch, cover lock, New All and Reset keys, code: 1..4 }
  $0000, $0000, $0000, $0000,
{ first row of small keys: F to Contrast, code: 5..14 }
  $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000,
{ second row of small keys: Q to P, code: 15..24 }
  $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000,
{ third row of small keys: A to L, code: 25..33 }
  $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000,
{ fourth row of small keys: Caps to Eng, code: 34..42 }
  $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000,
{ two SHIFT keys and the SPACE bar, code: 43..45 }
  $0000, $0000, $0080,
{ five rows of large keys, code: 46..68 }
  $0000, $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000, $0020,
  $0002, $0004, $0008,
{ EXE key, code: 69 }
  $0010,
{ row of 13 membrane keys below the LCD, code: 70..82 }
  $0000, $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000, $0000,
  $0000, $0000, $0000,
{ touch screen pads, code: 83..98 }
  $0000, $0000, $0000, $1000,
  $0000, $0000, $0000, $2000,
  $0000, $0000, $0000, $4000,
  $0000, $0000, $0000, $8000,
{ KANA key, code 99 }
  $0001
),

( { KO11 }
{ no key pressed }
  $0000,
{ power switch, cover lock, New All and Reset keys, code: 1..4 }
  $0000, $0000, $0000, $0000,
{ first row of small keys: F to Contrast, code: 5..14 }
  $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000,
{ second row of small keys: Q to P, code: 15..24 }
  $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000,
{ third row of small keys: A to L, code: 25..33 }
  $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000,
{ fourth row of small keys: Caps to Eng, code: 34..42 }
  $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000,
{ two SHIFT keys and the SPACE bar, code: 43..45 }
  $0040, $0040, $0080,
{ five rows of large keys, code: 46..68 }
  $0000, $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000, $0000,
  $0000, $0000, $0000,
{ EXE key, code: 69 }
  $0000,
{ row of 13 membrane keys below the LCD, code: 70..82 }
  $0000, $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000, $0000,
  $0000, $0000, $0000,
{ touch screen pads, code: 83..98 }
  $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000,
{ KANA key, code 99 }
  $0000
),

( { KO12 }
{ no key pressed }
  $0000,
{ power switch, cover lock, New All and Reset keys, code: 1..4 }
  $0000, $0000, $0000, $0000,
{ first row of small keys: F to Contrast, code: 5..14 }
  $0040, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000,
{ second row of small keys: Q to P, code: 15..24 }
  $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000,
{ third row of small keys: A to L, code: 25..33 }
  $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000,
{ fourth row of small keys: Caps to Eng, code: 34..42 }
  $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000,
{ two SHIFT keys and the SPACE bar, code: 43..45 }
  $0000, $0000, $0000,
{ five rows of large keys, code: 46..68 }
  $0000, $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000, $0000,
  $0000, $0000, $0000,
{ EXE key, code: 69 }
  $0000,
{ row of 13 membrane keys below the LCD, code: 70..82 }
  $0000, $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000, $0000,
  $0000, $0000, $0000,
{ touch screen pads, code: 83..98 }
  $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000,
{ KANA key, code 99 }
  $0000
),

( { KO code 13, all columns selected }
{ no key pressed }
  $0000,
{ power switch, cover lock, New All and Reset keys, code: 1..4 }
  $0000, $0000, $0004, $0000,
{ first row of small keys: F to Contrast, code: 5..14 }
  $0040, $0020, $0010, $0008, $0004, $0002, $0001, $0080, $0080, $0080,
{ second row of small keys: Q to P, code: 15..24 }
  $0020, $0010, $0008, $0004, $0002, $0001, $0080, $0080, $0001, $0002,
{ third row of small keys: A to L, code: 25..33 }
  $0010, $0008, $0004, $0002, $0001, $0080, $0080, $0001, $0002,
{ fourth row of small keys: Caps to Eng, code: 34..42 }
  $0020, $0020, $0010, $0008, $0004, $0002, $0001, $0080, $0001,
{ two SHIFT keys and the SPACE bar, code: 43..45 }
  $0040, $0040, $0080,
{ five rows of large keys, code: 46..68 }
  $0080, $0001, $0002, $0008, $0010,
  $0002, $0004, $0008, $0010, $0020,
  $0004, $0008, $0010, $0020, $0020,
  $0004, $0008, $0010, $0020, $0020,
  $0002, $0004, $0008,
{ EXE key, code: 69 }
  $0010,
{ row of 13 membrane keys below the LCD, code: 70..82 }
  $2000, $2000, $2000, $2000, $2000,
  $4000, $4000, $4000, $4000, $8000,
  $8000, $8000, $8000,
{ touch screen pads, code: 83..98 }
  $1000, $1000, $1000, $1000,
  $2000, $2000, $2000, $2000,
  $4000, $4000, $4000, $4000,
  $8000, $8000, $8000, $8000,
{ KANA key, code 99 }
  $0001
),

( { KO code 14, undefined }
{ no key pressed }
  $0000,
{ power switch, cover lock, New All and Reset keys, code: 1..4 }
  $0000, $0000, $0000, $0000,
{ first row of small keys: F to Contrast, code: 5..14 }
  $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000,
{ second row of small keys: Q to P, code: 15..24 }
  $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000,
{ third row of small keys: A to L, code: 25..33 }
  $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000,
{ fourth row of small keys: Caps to Eng, code: 34..42 }
  $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000,
{ two SHIFT keys and the SPACE bar, code: 43..45 }
  $0000, $0000, $0000,
{ five rows of large keys, code: 46..68 }
  $0000, $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000, $0000,
  $0000, $0000, $0000,
{ EXE key, code: 69 }
  $0000,
{ row of 13 membrane keys below the LCD, code: 70..82 }
  $0000, $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000, $0000,
  $0000, $0000, $0000,
{ touch screen pads, code: 83..98 }
  $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000,
{ KANA key, code 99 }
  $0000
),

( { KO code 15, undefined }
{ no key pressed }
  $0000,
{ power switch, cover lock, New All and Reset keys, code: 1..4 }
  $0000, $0000, $0000, $0000,
{ first row of small keys: F to Contrast, code: 5..14 }
  $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000,
{ second row of small keys: Q to P, code: 15..24 }
  $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000,
{ third row of small keys: A to L, code: 25..33 }
  $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000,
{ fourth row of small keys: Caps to Eng, code: 34..42 }
  $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000,
{ two SHIFT keys and the SPACE bar, code: 43..45 }
  $0000, $0000, $0000,
{ five rows of large keys, code: 46..68 }
  $0000, $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000, $0000,
  $0000, $0000, $0000,
{ EXE key, code: 69 }
  $0000,
{ row of 13 membrane keys below the LCD, code: 70..82 }
  $0000, $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000, $0000,
  $0000, $0000, $0000,
{ touch screen pads, code: 83..98 }
  $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000,
  $0000, $0000, $0000, $0000,
{ KANA key, code 99 }
  $0000
)

);


function ReadKy (Ko: byte) : word;
begin
  ReadKy := KeyTab[Ko,KeyCode1] or KeyTab[Ko,KeyCode2];
end {ReadKy};


end.
