unit Main;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, StdCtrls, IniFiles, ThdTimer;

type
  TMainForm = class(TForm)
    RefreshTimer: TTimer;
    CursorTimer: TTimer;
    RunTimer: TThreadedTimer;
    WatchTimer: TTimer;
    procedure OnRefreshTimer(Sender: TObject);
    procedure OnCursorTimer(Sender: TObject);
    procedure OnRunTimer(Sender: TObject);
    procedure FormMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure FormMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure FormMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure FormShow(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormClose(Sender: TObject; var CloseAction: TCloseAction);
    procedure FormKeyPress(Sender: TObject; var Key: Char);
    procedure FormKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure FormKeyUp(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure FormPaint(Sender: TObject);
    procedure FormDeactivate(Sender: TObject);
    procedure ApplicationDeactivate(Sender: TObject);
    procedure OnWatchTimer(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
  end;

var
    MainForm: TMainForm;

implementation

{$R *.dfm}

uses
    Def, Cpu, Keyboard, Debug;

const
    FaceName: string = 'face.bmp';
    DigitsName: string = 'digits.bmp';
    RomName: string = 'rom.bin';
    RamName: string = 'ram.bin';
    LoadMsg: string = 'Failed to load the file ';
    SaveMsg: string = 'Failed to save the file ';

var
    BitMap, FaceBmp, LcdBmp, HlpBmp, TimeBmp, DigitsBmp: TBitMap;
    RedrawReq: boolean;		{ true if the LcdBmp image has changed and needs
				  to be redrawn }
{ watch }
    WatchMode: integer;
    WatchData: array[0..6] of integer = ( -1, -1, -1, -1, -1, -1, -1 );
    WatchDot: boolean = True;

{ LCD }
    CurVis: Boolean = False;
    Scr: array[0..95] of byte;	{ LCD shadow memory }
    clLcdBkg: TColor;		{ LCD background }
    Digits7seg: array[1..4] of byte = ( $FF, $FF, $FF, $FF );

{ CPU }
    CpuSpeed: integer;		{ how many instructions executes the emulated
				  CPU at each RunTimer event call }
    ic: integer = 0;		{ counter of executed instructions }


{ display a status sign }
procedure ViewHlp (
	x: integer;
	y: integer;	{ coordinates of the upper-left corner }
        w: integer;	{ image width }
        h: integer;	{ image height }
	what: byte);	{ draw image if not zero, clear if zero }
begin
  if what <> 0 then
  begin		{ copy a selection from the HlpBmp }
    with BitMap do
    begin
      Transparent := False;
      Width := w;
      Height := h;
      Canvas.Draw (-x, -y, HlpBmp);
    end {with};
    LcdBmp.Canvas.Draw (x, y, BitMap);
  end
  else
  begin		{ clear the area }
    with LcdBmp.Canvas do
    begin
      Brush.Style := bsSolid;
      Brush.Color := clLcdBkg;
      FillRect (Rect(x, y, x+w, y+h));
    end {with};
  end {if};
  RedrawReq := True;
end {ViewHlp};


{ display a 7-segment character }
procedure View7seg (
	DstX: integer;	{ Y coordinate on the LcdBmp }
	pattern: byte);	{ segments bit pattern: *gfedcba }
const
  DstY = 0;		{ Y coordinate on the LcdBmp }
  SrcY = 44;		{ Y coordinate on the DigitsBmp }
  W = 15;		{ width of the character }
  W1 = W div 2;		{ width of the middle left part }
  W2 = W - W1;		{ width of the middle right part }
  H1 = 7;		{ height of the upper part }
  H2 = 7;		{ height of the middle parts }
  H3 = 6;		{ height of the lower part }
  Tab1: array[0..7] of integer = ( 7*W, 5*W, 3*W, 1*W, 6*W, 4*W, 2*W, 0*W );
  Tab2: array[0..7] of integer = ( 11*W, 9*W, 10*W, 8*W, 5*W, 1*W, 4*W, 0*W );
  Tab3: array[0..7] of integer = ( 7*W, 6*W, 5*W, 4*W, 3*W, 2*W, 1*W, 0*W );
var
  SrcX: integer;
  p: cardinal;
begin
{ upper part: half of f, half of b, whole a }
  p := cardinal(pattern) and $23;
  if p >= $20 then Dec (p, $1C);
  SrcX := Tab1[p];
  with BitMap do
  begin
    Transparent := False;
    Width := W;
    Height := H1;
    Canvas.Draw (-SrcX, -SrcY, DigitsBmp);
  end {with};
  LcdBmp.Canvas.Draw (DstX, DstY, BitMap);
{ middle left part: half of g, half of f, half of e }
  p := (cardinal(pattern) and $70) shr 4;
  SrcX := Tab2[p];
  with BitMap do
  begin
    Transparent := False;
    Width := W1;
    Height := H2;
    Canvas.Draw (-SrcX, -SrcY-H1, DigitsBmp);
  end {with};
  LcdBmp.Canvas.Draw (DstX, DstY+H1, BitMap);
{ middle right part: half of g, half of c, half of b }
  p := (cardinal(pattern) and $46) shr 1;
  if p >= $20 then Dec (p, $1C);
  SrcX := Tab2[p];
  with BitMap do
  begin
    Transparent := False;
    Width := W2;
    Height := H2;
    Canvas.Draw (-SrcX-W1, -SrcY-H1, DigitsBmp);
  end {with};
  LcdBmp.Canvas.Draw (DstX+W1, DstY+H1, BitMap);
{ lower part: half of e, whole d, half of c }
  p := (cardinal(pattern) and $1C) shr 2;
  SrcX := Tab3[p];
  with BitMap do
  begin
    Transparent := False;
    Width := W;
    Height := H3;
    Canvas.Draw (-SrcX, -SrcY-H1-H2, DigitSBmp);
  end {with};
  LcdBmp.Canvas.Draw (DstX, DstY+H1+H2, BitMap);
  RedrawReq := True;
end {View7seg};


{ draw the LCD contents }
procedure View;
var
  Pos, Row, Col, Index, ZX, ZY: Integer;
  B: byte;
begin

{ draw the matrix }
  Index := 0;
  ZX := 0;	{ pixel coordinates }
  ZY := 32;

  with LcdBmp.Canvas do
  begin
    Brush.Style := bsSolid;
    for Pos := 0 to 11 do
    begin
      Inc (Index);
      for Row := 1 to 7 do
      begin
        B := lcd[Index];

{ handle the cursor }
        if CurVis and (Pos = (lcd[96] and $0F)) then
          if (Row = 7) or ((lcd[96] and $10) = 0) then B := $1F else B := 0;

        if Scr[Index] <> B then
        begin
          RedrawReq := True;
          Scr[Index] := B;
          for Col := 0 to 4 do
          begin
            if (B and 1) <> 0 then Brush.Color := clBlack
                              else Brush.Color := clLcdBkg;
            B := B shr 1;
            FillRect (Rect(ZX, ZY, ZX+2, ZY+3));
            Inc (ZX, 3);
          end {for Col};
          Dec (ZX, 15);
        end {if};
        Inc (Index);
        Inc (ZY, 4);
      end {for Row};
      Inc (ZX, 22);
      Dec (ZY, 4*7)
    end {for Pos};
  end {with};

{ draw the 7-segm. characters and the status signs }
  if Scr[$00] <> lcd[$00] then
  begin
    Scr[$00] := lcd[$00];
    ViewHlp (2, 10, 12, 12, lcd[$00] and $01);		{ latin F }
    ViewHlp (3, 0, 15, 11, lcd[$00] and $02);		{ <- }
 end {if};

  if Scr[$08] <> lcd[$08] then
  begin
    Scr[$08] := lcd[$08];
    ViewHlp (19, 8, 8, 7, lcd[$08] and $02);		{ down }
    ViewHlp (19, 15, 8, 7, lcd[$08] and $04);		{ up }
    ViewHlp (31, 11, 21, 11, lcd[$08] and $08);		{ RANG }
    ViewHlp (34, 0, 14, 11, lcd[$08] and $10);		{ VV }
  end {if};

  if Scr[$10] <> lcd[$10] then
  begin
    Scr[$10] := lcd[$10];
    ViewHlp (55, 0, 15, 11, lcd[$10] and $02);		{ key }
    ViewHlp (55, 11, 15, 11, lcd[$10] and $04);		{ tel }
    ViewHlp (72, 0, 15, 12, lcd[$10] and $08);		{ table }
    ViewHlp (72, 11, 15, 11, lcd[$10] and $10);		{ notes }
  end {if};

  if Scr[$18] <> lcd[$18] then
  begin
    Scr[$18] := lcd[$18];
    ViewHlp (90, 0, 22, 11, lcd[$18] and $08);		{ AVTO }
    ViewHlp (90, 11, 22, 12, lcd[$18] and $10);		{ RED }
  end {if};

  if Scr[$20] <> lcd[$20] then
  begin
    Scr[$20] := lcd[$20];
    ViewHlp (114, 0, 9, 10, lcd[$20] and $08);		{ memory }
    ViewHlp (114, 10, 9, 10, lcd[$20] and $10);		{ K }
  end {if};

  if Scr[$28] <> lcd[$28] then
  begin
    Scr[$28] := lcd[$28];
    ViewHlp (125, 0, 10, 10, lcd[$28] and $02); 	{ div }
    ViewHlp (125, 10, 10, 10, lcd[$28] and $04); 	{ mul }
    ViewHlp (138, 0, 10, 10, lcd[$28] and $08); 	{ - }
    ViewHlp (138, 10, 10, 10, lcd[$28] and $10); 	{ + }
  end {if};

  if Scr[$50] <> lcd[$50] then
  begin
    Scr[$50] := lcd[$50];
    ViewHlp (209, 3, 9, 20, lcd[$50] and $02);		{ "1" }
  end {if};

  if Scr[$58] <> lcd[$58] then
  begin
    Scr[$58] := lcd[$58];
    ViewHlp (241, 0, 15, 11, lcd[$58] and $10);		{ -> }
  end {if};

  B := 0;
  if (lcd[$30] and $02) <> 0 then Inc (B, $40);		{ 1g }
  if (lcd[$30] and $01) <> 0 then Inc (B, $20);		{ 1f }
  if (lcd[$30] and $04) <> 0 then Inc (B, $10);		{ 1e }
  if (lcd[$30] and $08) <> 0 then Inc (B, $08);		{ 1d }
  if (lcd[$30] and $10) <> 0 then Inc (B, $04);		{ 1c }
  if (lcd[$38] and $01) <> 0 then Inc (B, $02);		{ 1b }
  if (lcd[$38] and $02) <> 0 then Inc (B, $01);		{ 1a }
  if Digits7seg[1] <> B then
  begin
    Digits7seg[1] := B;
    View7seg (153, B);
  end {if};

  B := 0;
  if (lcd[$38] and $08) <> 0 then Inc (B, $40);		{ 2g }
  if (lcd[$38] and $04) <> 0 then Inc (B, $20);		{ 2f }
  if (lcd[$38] and $10) <> 0 then Inc (B, $10);		{ 2e }
  if (lcd[$40] and $01) <> 0 then Inc (B, $08);		{ 2d }
  if (lcd[$40] and $02) <> 0 then Inc (B, $04);		{ 2c }
  if (lcd[$40] and $04) <> 0 then Inc (B, $02);		{ 2b }
  if (lcd[$40] and $08) <> 0 then Inc (B, $01);		{ 2a }
  if Digits7seg[2] <> B then
  begin
    Digits7seg[2] := B;
    View7seg (172, B);
  end {if};

  B := 0;
  if (lcd[$48] and $01) <> 0 then Inc (B, $40);		{ 3g }
  if (lcd[$40] and $10) <> 0 then Inc (B, $20);		{ 3f }
  if (lcd[$48] and $02) <> 0 then Inc (B, $10);		{ 3e }
  if (lcd[$48] and $04) <> 0 then Inc (B, $08);		{ 3d }
  if (lcd[$48] and $08) <> 0 then Inc (B, $04);		{ 3c }
  if (lcd[$48] and $10) <> 0 then Inc (B, $02);		{ 3b }
  if (lcd[$50] and $01) <> 0 then Inc (B, $01);		{ 3a }
  if Digits7seg[3] <> B then
  begin
    Digits7seg[3] := B;
    View7seg (191, B);
  end {if};

  B := 0;
  if (lcd[$50] and $08) <> 0 then Inc (B, $40);		{ 4g }
  if (lcd[$50] and $04) <> 0 then Inc (B, $20);		{ 4f }
  if (lcd[$50] and $10) <> 0 then Inc (B, $10);		{ 4e }
  if (lcd[$58] and $01) <> 0 then Inc (B, $08);		{ 4d }
  if (lcd[$58] and $02) <> 0 then Inc (B, $04);		{ 4c }
  if (lcd[$58] and $04) <> 0 then Inc (B, $02);		{ 4b }
  if (lcd[$58] and $08) <> 0 then Inc (B, $01);		{ 4a }
  if Digits7seg[4] <> B then
  begin
    Digits7seg[4] := B;
    View7seg (219, B);
  end {if};
end; {proc View}


{ In order to avoid display flickers all drawing is done off-screen
  on LcdBmp.Canvas, then periodically transferred to MainForm.Canvas }
procedure TMainForm.OnRefreshTimer(Sender: TObject);
begin
  View;
  if RedrawReq = True then Canvas.Draw (70, 110, LcdBmp);
  RedrawReq := False;
end;


{ release a pressed key if it's placed outside the coordinates X,Y }
procedure ReleaseKey1 (X, Y: Integer);
var
  i, r, c, k: integer;
begin
  if KeyCode1 = 0 then Exit;

{ locate the "keyblock" the key "KeyCode1" belongs to }
  i := 0;	{ "keyblock" index }
  k := 1;	{ first key code in the "keyblock" }
  while (KeyCode1 >= k + keypad[i].cnt) and (i < KEYPADS) do
  begin
    Inc (k, keypad[i].cnt);
    Inc (i);
  end {while};

  with keypad[i] do
  begin
    k := KeyCode1 - k;		{ offset of the key in the "keyblock" }
    c := L + SX*(k mod col);	{ X coordinate of the key image }
    r := T + SY*(k div col);	{ Y coordinate of the key image }
    if (X < c) or (X >= c + W) or (Y < r) or (Y >= r + H) then KeyCode1 := 0;
  end {with};
end {ReleaseKey1};


{ called when mouse button pressed }
procedure TMainForm.FormMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var
  i, r, c, k: Integer;
begin
{ proceed only when left mouse button pressed }
  if Button <> mbLeft then Exit;

  ReleaseKey1 (-1, -1);
  KeyCode1 := 1;
  for i := 0 to KEYPADS do
  begin
    with keypad[i] do
    begin
      if (X >= L) and (X < L+SX*col) and (((X-L) mod SX) < W) and
	(Y >= T) and (((Y-T) mod SY) < H) then
      begin
        c := (X-L) div SX;
        r := (Y-T) div SY;
        k := col*r + c;
        if k < cnt then
        begin
          Inc (KeyCode1, k);
          break;
        end {if};
      end {if};
      Inc (KeyCode1, cnt);
    end {with};
  end {for};

  if KeyCode1 > LASTKEYCODE then	{ no valid key pressed }
  begin
    KeyCode1 := 0;
{ dragging a captionless form by clicking anywhere on the client area outside
  the controls }
    if BorderStyle = bsNone then
    begin
      ReleaseCapture;
      SendMessage (Handle, WM_NCLBUTTONDOWN, HTCAPTION, 0);
    end {if};
  end {if};
end {proc};


{ called when mouse button released }
procedure TMainForm.FormMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var
  K: integer;
begin
{ proceed only when left mouse button was pressed }
  if Button <> mbLeft then Exit;

  K := KeyCode1;
{ release a pressed key }
  ReleaseKey1 (-1, -1);

{ what to do if the mouse button was released over a pressed ... }
  case K of
    76: HALT_i := true;		{ ...power on button }
    77: Close;			{ ...power off button }
  end {case};
end;


{ called when moving the mouse while the button pressed }
procedure TMainForm.FormMouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
begin
{ release a pressed key if mouse was moved from it }
  ReleaseKey1 (X, Y);
end;


procedure TMainForm.FormShow(Sender: TObject);
var
  X: Integer;
begin
  KeyCode1 := 0;
  KeyCode2 := 0;
  CpuStop := False;
  CpuDelay := 0;
  CpuSteps := -1;
  BreakPoint := -1;
{ load the DigitsBmp image }
  if FileExists (DigitsName) then
    DigitsBmp.LoadFromFile (DigitsName)
  else
    MessageDlg (LoadMsg + DigitsName, mtWarning, [mbOk], 0);
  DigitsBmp.Transparent := False;
{ draw the background image on the FaceBmp.Canvas }
  if FileExists (FaceName) then
  begin
    FaceBmp.LoadFromFile (FaceName);
    FaceBmp.Transparent := False;
    HlpBmp.Canvas.Draw (-70, -110, FaceBmp);
    HlpBmp.Transparent := False;
    TimeBmp.Canvas.Draw (-255, -237, FaceBmp);
    TimeBmp.Transparent := False;
    MainForm.Canvas.Draw (0, 0, FaceBmp);
    clLcdBkg := FaceBmp.Canvas.Pixels [68, 108];
  end
  else
  begin
    MessageDlg (LoadMsg + FaceName, mtWarning, [mbOk], 0);
    clLcdBkg := clWhite;
  end {if};
{ clear the LCD area }
  with LcdBmp.Canvas do
  begin
    Brush.Style := bsSolid;
    Brush.Color := clLcdBkg;
    FillRect (Rect(0, 0, 257, 60));
  end {with};
{ clear the display memory }
  for X := 0 to 95 do
  begin
    Scr[X] := $FF;
    lcd[X] := $00;
  end {for};
  CpuReset;
  RunTimer.Enabled := True;
  RefreshTimer.Enabled := True;
  CursorTimer.Enabled := True;
  WatchTimer.Enabled := (WatchMode >= 0) and (WatchMode <= 2);
  RedrawReq := True;
end;


{ load the ROM and RAM images }
procedure MemLoad;
var
  f: file;
  transferred: integer;
begin
{ load the ROM image }
  if FileExists (RomName) then
  begin
    AssignFile (f, RomName);
    Reset (f, 1);
    BlockRead (f, rom, ROMSIZE, transferred);
    CloseFile (f);
  end
  else MessageDlg (LoadMsg + RomName, mtWarning, [mbOk], 0);
{ load the RAM image }
  if FileExists (RamName) then
  begin
    AssignFile (f, RamName);
    Reset (f, 1);
    BlockRead (f, ram, RamSize, transferred);
    CloseFile (f);
  end;
end {MemLoad};


procedure IniLoad;
var
  IniMK: TIniFile;
begin
  IniMK := TIniFile.Create (ExpandFileName(IniName));
  with IniMK do
  begin
    CpuSpeed := ReadInteger ('Settings', 'CpuSpeed', 1000);
    RamSize := ReadInteger ('Settings', 'RamSize', 2048);
    WatchMode := ReadInteger ('Settings', 'WatchMode', 0);
  end {with};
  IniMK.Free;
end {IniLoad};


{ initialise the application }
procedure TMainForm.FormCreate(Sender: TObject);
begin
  BitMap := TBitMap.Create;
  FaceBmp := TBitMap.Create;
  FaceBmp.Width := 747;
  FaceBmp.Height := 700;
  LcdBmp := TBitMap.Create;
  LcdBmp.Width := 257;
  LcdBmp.Height := 60;
  HlpBmp := TBitMap.Create;
  HlpBmp.Width := 257;
  HlpBmp.Height := 23;
  TimeBmp := TBitMap.Create;
  TimeBmp.Width := 77;
  TimeBmp.Height := 51;
  DigitsBmp := TBitMap.Create;
  DigitsBmp.Width := 187;
  DigitsBmp.Height := 64;
  IniLoad;
  if RamSize < MINRAMSIZE then RamSize := MINRAMSIZE;
  if RamSize > MAXRAMSIZE then RamSize := MAXRAMSIZE;
  RamSize := (RamSize + 15) and $FFF0;
  RamEnd := RAMSTART + RamSize - 1;
  MemLoad;
end;


{ save the RAM image to the file Ram.bin }
procedure MemSave;
var
  f: file;
begin
  {$I-}
  AssignFile (f, RamName);
  Rewrite (f, 1);
  BlockWrite (f, ram, RamSize);
  CloseFile (f);
  {$I+}
  if IOResult <> 0 then MessageDlg (SaveMsg + RamName, mtWarning, [mbOk], 0);
end {MemSave};


{ terminate the application }
procedure TMainForm.FormClose(Sender: TObject; var CloseAction: TCloseAction);
begin
  CpuStop := True;
  RunTimer.Enabled := False;
  RefreshTimer.Enabled := False;
  CursorTimer.Enabled := False;
  WatchTimer.Enabled := False;
  MemSave;
  BitMap.Free;
  FaceBmp.Free;
  LcdBmp.Free;
  HlpBmp.Free;
  TimeBmp.Free;
  DigitsBmp.Free;
end;


procedure TMainForm.OnCursorTimer(Sender: TObject);
begin
  CurVis := not CurVis;
end;


procedure TMainForm.FormKeyPress(Sender: TObject; var Key: Char);
const
{ Key codes $20..$3F }
  tab1: array[0..63] of byte = (
    73, 0,  48, 1,  49, 1,  0, 0, { space ! " }
    51, 1,  15, 0,  00, 0,  45, 1, { $ % & ' }
    41, 1,  42, 1,  24, 0,  34, 0, { ( ) * + }
    44, 1,  29, 0,  32, 0,  19, 0, { , - . / }
    31, 0,  26, 0,  27, 0,  28, 0, { 0 1 2 3 }
    21, 0,  22, 0,  23, 0,  16, 0, { 4 5 6 7 }
    17, 0,  18, 0,  43, 1,  53, 1, { 8 9 : ; }
    54, 1,  33, 0,  55, 1,  56, 1 { < = > ? }
  );
{ Key codes $E0..$FF - lower case cyrillic letters }
  tab2: array[0..63] of byte = (
    41, 0,  42, 0,  43, 0,  44, 0, { a b v g }
    45, 0,  46, 0,  48, 0,  49, 0, { d e zh z }
    50, 0,  51, 0,  52, 0,  53, 0, { i j k l }
    54, 0,  55, 0,  56, 0,  57, 0, { m n o p }
    58, 0,  59, 0,  60, 0,  61, 0, { r s t u }
    62, 0,  63, 0,  64, 0,  65, 0, { f h c ch }
    66, 0,  67, 0,  57, 1,  68, 0, { sh shh tvjordyj_znak y }
    69, 0,  70, 0,  71, 0,  72, 0  { mjagkij_znak je ju ja }
  );
var
  i: integer;
begin
  i := integer(cardinal(byte(Key)));
  case i of
    $0E: KeyCode2 := 2;		{ N = notepad (memory) }
    $10: KeyCode2 := 1;		{ P = phone book }
    $14: KeyCode2 := 6;		{ T = table }
    $1B: KeyCode2 := 12;	{ Escape = cyrillic "S" }
    $20..$3F: begin
        KeyCode2 := integer (tab1[(i - $20) * 2]);
        ram[$70] := ram[$70] or tab1[(i - $20) * 2 + 1];
      end;
    $A8, $B8: KeyCode2 := 47;	{ jo }
    $C0..$DF: begin		{ upper case cyrillic letters }
        KeyCode2 := integer (tab2[(i - $C0) * 2]);
        ram[$70] := ram[$70] or tab2[(i - $C0) * 2 + 1];
      end;
    $E0..$FF: begin		{ lower case cyrillic letters }
        KeyCode2 := integer (tab2[(i - $E0) * 2]);
        ram[$70] := ram[$70] or tab2[(i - $E0) * 2 + 1];
      end;
  end {case};
end;


procedure TMainForm.FormKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  case Key of
    $BB:			{ + }
      if ssCtrl in Shift then KeyCode2 := 35;
    $BD:			{ - }
      if ssCtrl in Shift then KeyCode2 := 30;
    VK_LEFT:			{ arrow left }
      if ssCtrl in Shift then KeyCode2 := 9 else KeyCode2 := 4;
    VK_RIGHT:			{ arrow right }
      if ssCtrl in Shift then KeyCode2 := 10 else KeyCode2 := 5;
    VK_UP:			{ arrow up }
      if ssCtrl in Shift then KeyCode2 := 3;
    VK_DOWN:			{ arrow down }
      if ssCtrl in Shift then KeyCode2 := 8;
    VK_DELETE:			{ cyrillic "SP" / "SK" }
      if ssCtrl in Shift then KeyCode2 := 20 else KeyCode2 := 13;
    VK_RETURN:			{ M->x / diamond }
      if ssCtrl in Shift then KeyCode2 := 25 else KeyCode2 := 75;
    VK_INSERT:	KeyCode2 := 38;	{ cyrillic "VN" }
    VK_BACK:	KeyCode2 := 39;	{ cyrillic "UD" }
    VK_HOME:	KeyCode2 := 36;	{ cyrillic "VV" }
    VK_END:	KeyCode2 := 40;	{ cyrillic "SI" }
    VK_F3:	DebugForm.Show;
    VK_F8:	KeyCode2 := 78;	{ Init }
    VK_F7:      KeyCode2 := 79; { reveal password }
    VK_F6:      KeyCode2 := 80; { information about the author }
    VK_F5:      KeyCode2 := 81;
  end {case};
end;


procedure TMainForm.FormKeyUp(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  KeyCode2 := 0;
end;


procedure TMainForm.FormPaint(Sender: TObject);
begin
  FaceBmp.Canvas.Draw (70, 110, LcdBmp);
  FaceBmp.Canvas.Draw (255, 237, TimeBmp);
  Canvas.Draw (0, 0, FaceBmp);
  RedrawReq := False;
end;


{ execute a bunch of machine code instructions }
procedure TMainForm.OnRunTimer(Sender: TObject);
begin
  if CpuDelay > 0 then
  begin
    Dec (CpuDelay);
    Exit;
  end {if};

  Inc (ic, CpuSpeed);
  while ic > 0 do
  begin
    if CpuStop then
    begin
      ic := 0;
      break;
    end {if};

    Dec (ic, CpuRun);
    if CpuSteps > 0 then
    begin
      Dec (CpuSteps);
      if CpuSteps = 0 then
      begin
        CpuStop := True;
        ic := 0;
        DebugForm.Show;
        break;
      end {if};
    end {if};

    if (BreakPoint >= 0) and (BreakPoint = ptrw(@reg[R7])^) then
    begin
      CpuStop := True;
      ic := 0;
      DebugForm.Show;
      break;
    end {if};
  end {while};
end;


procedure TMainForm.FormDeactivate(Sender: TObject);
begin
  ReleaseKey1 (-1, -1);
  KeyCode2 := 0;
end;


procedure TMainForm.ApplicationDeactivate(Sender: TObject);
begin
  ReleaseKey1 (-1, -1);
  KeyCode2 := 0;
end;


{ returns true if the TimeBmp image has changed and needs to be redrawn }
function TimeView: boolean;
const
{ X position of the digits on the TimeBmp }
  LargeX: array[0..3] of integer = ( 2, 19, 42, 59 );
  SmallX: array[0..1] of integer = ( 51, 64 );
{ X position of the day of the week dots on the TimeBmp }
  DaysX: array[1..7] of integer = ( 39, 0, 7, 13, 20, 26, 33 );
var
  t: TDateTime;
  i: integer;
  Hour, Min, Sec, MSec, Year, Month, Day: word;
  data: array[0..3] of integer;
  dot: boolean;
begin
  TimeView := False;
  t := Now;
  DecodeTime(t, Hour, Min, Sec, MSec);
  DecodeDate(t, Year, Month, Day);
{ large digits }
  case WatchMode of
    0, 1: begin
        data[0] := Hour div 10;
        if data[0] = 0 then data[0] := 10; { suppress the leading 0 }
        data[1] := Hour mod 10;
        data[2] := Min div 10;
        data[3] := Min mod 10;
      end;
    2: begin
        data[0] := Day div 10;
        if data[0] = 0 then data[0] := 10; { suppress the leading 0 }
        data[1] := Day mod 10;
        data[2] := Month div 10;
        if data[2] = 0 then data[2] := 10; { suppress the leading 0 }
        data[3] := Month mod 10;
      end;
    else Exit;
  end {case};
  for i := 0 to 3 do
  begin
    if WatchData[i] <> data[i] then
    begin
      with BitMap do
      begin
        Width := 17;
        Height := 25;
        Canvas.Draw (-17*data[i], 0, DigitsBmp);
        Transparent := False;
      end {with};
      TimeBmp.Canvas.Draw (LargeX[i], 0, BitMap);
      WatchData[i] := data[i];
      TimeView := True;
    end {if};
  end {for};
{ small digits }
  case WatchMode of
    0: begin
        data[0] := Sec div 10;
        data[1] := Sec mod 10;
        dot := MSec > 500;
      end;
    1: begin
        data[0] := Month div 10;
        if data[0] = 0 then data[0] := 10; { suppress the leading 0 }
        data[1] := Month mod 10;
        dot := True;
      end;
    2: begin
        Year := Year mod 100;
        data[0] := Year div 10;
        data[1] := Year mod 10;
        dot := False;
      end;
    else Exit;
  end {case};
  for i := 0 to 1 do
  begin
    if WatchData[i+4] <> data[i] then
    begin
      with BitMap do
      begin
        Width := 12;
        Height := 19;
        Canvas.Draw (-12*data[i], -25, DigitsBmp);
        Transparent := False;
      end {with};
      TimeBmp.Canvas.Draw (SmallX[i], 30, BitMap);
      WatchData[i+4] := data[i];
      TimeView := True;
    end {if};
  end {for};
{ flashing dot }
  if WatchDot <> dot then
  begin
    if dot then i := -133 else i := -127;
    with BitMap do
    begin
      Width := 6;
      Height := 7;
      Canvas.Draw (i, -26, DigitsBmp);
      Transparent := False;
    end {with};
    TimeBmp.Canvas.Draw (35, 19, BitMap);
    WatchDot := dot;
    TimeView := True;
  end {if};
{ day of the week dot }
  i := DayOfWeek(t);
  if WatchData[6] <> i then
  begin
    with TimeBmp.Canvas do
    begin
      Brush.Style := bsSolid;
      Brush.Color := clLcdBkg;
      FillRect (Rect(3, 42, 49, 49));
    end {with};
    with BitMap do
    begin
      Width := 7;
      Height := 7;
      Canvas.Draw (-133 - DaysX[i], -35, DigitsBmp);
      Transparent := False;
    end {with};
    TimeBmp.Canvas.Draw (3 + DaysX[i], 42, BitMap);
    WatchData[6] := i;
    TimeView := True;
  end {if};
end;


procedure TMainForm.OnWatchTimer(Sender: TObject);
begin
  if TimeView then Canvas.Draw (255, 237, TimeBmp);
end;

end.
