unit Main;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ScktComp, ExtCtrls, StdCtrls, Menus, IniFiles, ComCtrls;

type
  TMainForm = class(TForm)
    FddSocket: TServerSocket;
    MenuBevel: TBevel;
    MainMenu1: TMainMenu;
    File1: TMenuItem;
    Exit1: TMenuItem;
    Help1: TMenuItem;
    About1: TMenuItem;
    ConnectedBox: TGroupBox;
    ConnectedShape: TShape;
    TransferBox: TGroupBox;
    TransferShape: TShape;
    TransferTimer: TTimer;
    OpenDialog1: TOpenDialog;
    ButtonBox: TGroupBox;
    Button1: TButton;
    StatusBar1: TStatusBar;
    procedure FormCreate(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure OnTransferTimer(Sender: TObject);
    procedure FddClientRead(Sender: TObject; Socket: TCustomWinSocket);
    procedure ClientConnected(Sender: TObject; Socket: TCustomWinSocket);
    procedure ClientDisconnected(Sender: TObject; Socket: TCustomWinSocket);
    procedure OnMenuExit(Sender: TObject);
    procedure OnMenuAbout(Sender: TObject);
    procedure Button1Click(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
  end;

var
  MainForm: TMainForm;


implementation

{$R *.dfm}

uses Fdd, Bios;

const
  IniName = 'md100srv.ini';
  InsertStr = 'Insert';
  EjectStr = 'Eject';

var
  logname: string;
  logfile: TextFile;
  NewTransferColor: TColor = clBtnFace;


procedure IniLoad;
var
  Ini1: TIniFile;
begin
  Ini1 := TIniFile.Create (ExpandFileName(IniName));
  with Ini1 do
  begin
    MainForm.FddSocket.Port := ReadInteger ('MD100', 'Port', 0);
    logname := ReadString ('MD100', 'LogFile', '');
  end {with};
  Ini1.Free;
end {IniLoad};


procedure TMainForm.FormCreate(Sender: TObject);
begin
  ForceCurrentDirectory := True;
  if ParamCount = 0 then
  begin
    diskname := '';
    Button1.Caption := InsertStr;
  end
  else
  begin
    diskname := ParamStr (1);
    StatusBar1.SimpleText := ExtractFileName (diskname);
    Button1.Caption := EjectStr;
  end {if};
  IniLoad;
  if logname <> '' then
  begin
    AssignFile (logfile, logname);
    Rewrite (logfile);
  end {if};
  FddOpen;
  FddSocket.Open;
  TransferTimer.Enabled := True;
end;


procedure TMainForm.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  FddSocket.Close;
  FddClose;
  if logname <> '' then CloseFile (logfile);
end;


procedure TMainForm.OnTransferTimer(Sender: TObject);
begin
  TransferShape.Brush.Color := NewTransferColor;
  NewTransferColor := clBtnFace;
end;


procedure TMainForm.FddClientRead(Sender: TObject; Socket: TCustomWinSocket);
var
  x, y: byte;
begin
  NewTransferColor := $0000C0FF;
  Socket.ReceiveBuf (x, 1);
  y := FddTransfer (x);
  Socket.SendBuf (y, 1);
  if logname <> '' then Writeln (logfile, IntToHex(x,2)+' '+IntToHex(y,2));
end;


procedure TMainForm.ClientConnected(Sender: TObject;
  Socket: TCustomWinSocket);
begin
  if logname <> '' then Writeln (logfile, 'connected');
  if FddSocket.Socket.ActiveConnections = 1 then
	ConnectedShape.Brush.Color := $0000FF00;
end;


procedure TMainForm.ClientDisconnected(Sender: TObject;
  Socket: TCustomWinSocket);
begin
  if logname <> '' then Writeln (logfile, 'disconnected');
  if FddSocket.Socket.ActiveConnections = 1 then
	ConnectedShape.Brush.Color := clBtnFace;
end;


procedure TMainForm.OnMenuExit(Sender: TObject);
begin
  Close;
end;


procedure TMainForm.OnMenuAbout(Sender: TObject);
begin
  MessageDlg ('MD-100 Emulator'#13#10
    + 'Version 15  2015/03/21'#13#10
    + 'http://www.pisi.com.pl/piotr433/index.htm',
    mtInformation, [mbOK], 0);
end;


procedure TMainForm.Button1Click(Sender: TObject);
begin
  if Button1.Caption = InsertStr then
  begin
    if OpenDialog1.Execute then
    begin
      diskname := OpenDialog1.FileName;
      StatusBar1.SimpleText := ExtractFileName (diskname);
      FddOpen;
      Button1.Caption := EjectStr;
    end {if};
  end
  else
  begin
    diskname := '';
    StatusBar1.SimpleText := '';
    FddClose;
    Button1.Caption := InsertStr;
  end {if};
end;

end.
